/*
 *
 *  $Id: pvrusb2-hdw-low.c 589 2005-09-11 21:59:55Z isely $
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2-hdw-low.h"
#include "pvrusb2-hdw-internal.h"
#include "pvrusb2-debug.h"

#define PVR2_CTL_WRITE_ENDPOINT  0x01
#define PVR2_CTL_READ_ENDPOINT   0x81

void pvr2_reset_ctl_endpoints(struct pvr2_hdw *hdw)
{
    if (!hdw->usb_dev) return;
    usb_settoggle(hdw->usb_dev, PVR2_CTL_WRITE_ENDPOINT & 0xf,
                  !(PVR2_CTL_WRITE_ENDPOINT & USB_DIR_IN), 0);
    usb_settoggle(hdw->usb_dev, PVR2_CTL_READ_ENDPOINT & 0xf,
                  !(PVR2_CTL_READ_ENDPOINT & USB_DIR_IN), 0);
    usb_clear_halt(hdw->usb_dev,
                   usb_rcvbulkpipe(hdw->usb_dev,
                                   PVR2_CTL_READ_ENDPOINT & 0x7f));
    usb_clear_halt(hdw->usb_dev,
                   usb_sndbulkpipe(hdw->usb_dev,
                                   PVR2_CTL_WRITE_ENDPOINT & 0x7f));
}

static void pvr2_ctl_write_complete(struct urb *urb, struct pt_regs *regs)
{
    struct pvr2_hdw *hdw = urb->context;
    hdw->ctl_write_pend_flag = 0;
    if (hdw->ctl_read_pend_flag) return;
    complete(&hdw->ctl_done);
}

static void pvr2_ctl_read_complete(struct urb *urb, struct pt_regs *regs)
{
    struct pvr2_hdw *hdw = urb->context;
    hdw->ctl_read_pend_flag = 0;
    if (hdw->ctl_write_pend_flag) return;
    complete(&hdw->ctl_done);
}

static void pvr2_ctl_timeout(unsigned long data)
{
    struct pvr2_hdw *hdw = (struct pvr2_hdw *)data;
    if (hdw->ctl_write_pend_flag || hdw->ctl_read_pend_flag) {
        hdw->ctl_timeout_flag = !0;
        if (hdw->ctl_write_pend_flag && hdw->ctl_write_urb) {
            usb_unlink_urb(hdw->ctl_write_urb);
        }
        if (hdw->ctl_read_pend_flag && hdw->ctl_read_urb) {
            usb_unlink_urb(hdw->ctl_read_urb);
        }
    }
}

int pvr2_send_request(struct pvr2_hdw *hdw,
                      void *write_data,unsigned int write_len,
                      void *read_data,unsigned int read_len)
{
    unsigned int idx;
    int status;
    struct timer_list timer;
    if (!hdw->ctl_lock_held) {
	pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "Attempted to execute control transfer without lock!!");
        status = -EINVAL;
        goto done;
    }
    if (!hdw->flag_ok) {
	pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "Attempted to execute control transfer when device not ok");
        return -EIO;
    }
    if (!(hdw->ctl_read_urb && hdw->ctl_write_urb)) {
	pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "Attempted to execute control transfer"
                   " when USB is disconnected");
        return -EIO;
    }

    /* Ensure that we have sane parameters */
    if (!write_data) write_len = 0;
    if (!read_data) read_len = 0;
    if (write_len > PVR2_CTL_BUFFSIZE) {
        pvr2_trace(
            PVR2_TRACE_ERROR_LEGS,
            "Attempted to execute %d byte control-write transfer (limit=%d)",
            write_len,PVR2_CTL_BUFFSIZE);
        return -EINVAL;
    }
    if (read_len > PVR2_CTL_BUFFSIZE) {
        pvr2_trace(
            PVR2_TRACE_ERROR_LEGS,
            "Attempted to execute %d byte control-read transfer (limit=%d)",
            write_len,PVR2_CTL_BUFFSIZE);
        return -EINVAL;
    }
    if ((!write_len) && (!read_len)) {
        pvr2_trace(
            PVR2_TRACE_ERROR_LEGS,
            "Attempted to execute null control transfer?");
        return -EINVAL;
    }

    hdw->cmd_debug_state = 1;
    if (write_len) {
        hdw->cmd_debug_code = ((unsigned char *)write_data)[0];
    } else {
        hdw->cmd_debug_code = 0;
    }
    hdw->cmd_debug_write_len = write_len;
    hdw->cmd_debug_read_len = read_len;

    /* Initialize common stuff */
    init_completion(&hdw->ctl_done);
    hdw->ctl_timeout_flag = 0;
    hdw->ctl_write_pend_flag = 0;
    hdw->ctl_read_pend_flag = 0;
    init_timer(&timer);
    timer.expires = jiffies + HZ*4;
    timer.data = (unsigned long)hdw;
    timer.function = pvr2_ctl_timeout;

    if (write_len) {
        hdw->cmd_debug_state = 2;
        /* Transfer write data to internal buffer */
        for (idx = 0; idx < write_len; idx++) {
            hdw->ctl_write_buffer[idx] = ((unsigned char *)write_data)[idx];
        }
        /* Initiate a write request */
        usb_fill_bulk_urb(hdw->ctl_write_urb,
                          hdw->usb_dev,
                          usb_sndbulkpipe(hdw->usb_dev,
                                          PVR2_CTL_WRITE_ENDPOINT),
                          hdw->ctl_write_buffer,
                          write_len,
                          pvr2_ctl_write_complete,
                          hdw);
        hdw->ctl_write_urb->transfer_flags |= URB_ASYNC_UNLINK;
        hdw->ctl_write_urb->actual_length = 0;
        hdw->ctl_write_pend_flag = !0;
        status = usb_submit_urb(hdw->ctl_write_urb,GFP_KERNEL);
        if (status < 0) {
	    pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                       "Failed to submit write-control URB status=%d",status);
            hdw->ctl_write_pend_flag = 0;
        }
    }

    if (read_len) {
        hdw->cmd_debug_state = 3;
        memset(hdw->ctl_read_buffer,0x43,read_len);
        /* Initiate a read request */
        usb_fill_bulk_urb(hdw->ctl_read_urb,
                          hdw->usb_dev,
                          usb_rcvbulkpipe(hdw->usb_dev,
                                          PVR2_CTL_READ_ENDPOINT),
                          hdw->ctl_read_buffer,
                          read_len,
                          pvr2_ctl_read_complete,
                          hdw);
        hdw->ctl_read_urb->transfer_flags |= URB_ASYNC_UNLINK;
        hdw->ctl_read_urb->actual_length = 0;
        hdw->ctl_read_pend_flag = !0;
        status = usb_submit_urb(hdw->ctl_read_urb,GFP_KERNEL);
        if (status < 0) {
	    pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                       "Failed to submit read-control URB status=%d",status);
            hdw->ctl_read_pend_flag = 0;
        }
    }

    /* Start timer */
    add_timer(&timer);

    /* Now wait for all I/O to complete */
    hdw->cmd_debug_state = 4;
    while (hdw->ctl_write_pend_flag || hdw->ctl_read_pend_flag) {
        wait_for_completion(&hdw->ctl_done);
    }
    hdw->cmd_debug_state = 5;

    /* Stop timer */
    del_timer_sync(&timer);

    hdw->cmd_debug_state = 6;
    status = 0;

    if (hdw->ctl_timeout_flag) {
        status = -ETIMEDOUT;
        goto done;
    }

    if (write_len) {
        /* Validate results of write request */
        if ((hdw->ctl_write_urb->status != 0) &&
            (hdw->ctl_write_urb->status != -ENOENT) &&
            (hdw->ctl_write_urb->status != -ESHUTDOWN) &&
            (hdw->ctl_write_urb->status != -ECONNRESET)) {
            /* USB subsystem is reporting some kind of failure on the write */
            status = hdw->ctl_write_urb->status;
	    pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                       "control-write URB failure, status=%d",status);
            goto done;
        }
        if (hdw->ctl_write_urb->actual_length < write_len) {
            /* Failed to write enough data */
            status = -EIO;
	    pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                       "control-write URB short, expected=%d got=%d",
                       write_len,hdw->ctl_write_urb->actual_length);
            goto done;
        }
    }
    if (read_len) {
        /* Validate results of read request */
        if ((hdw->ctl_read_urb->status != 0) &&
            (hdw->ctl_read_urb->status != -ENOENT) &&
            (hdw->ctl_read_urb->status != -ESHUTDOWN) &&
            (hdw->ctl_read_urb->status != -ECONNRESET)) {
            /* USB subsystem is reporting some kind of failure on the read */
            status = hdw->ctl_read_urb->status;
	    pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                       "control-read URB failure, status=%d",status);
            goto done;
        }
        if (hdw->ctl_read_urb->actual_length < read_len) {
            /* Failed to read enough data */
            status = -EIO;
	    pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                       "control-read URB short, expected=%d got=%d",
                       read_len,hdw->ctl_read_urb->actual_length);
            goto done;
        }
        /* Transfer retrieved data out from internal buffer */
        for (idx = 0; idx < read_len; idx++) {
            ((unsigned char *)read_data)[idx] = hdw->ctl_read_buffer[idx];
        }
    }

 done:
    hdw->cmd_debug_state = 0;
    if (status < 0) {
        pvr2_hdw_render_useless_unlocked(hdw);
    }
    return status;
}

/*
 * pvr2_send_sequence().
 *
 * executes a sequence of commands ended by a zero length one.
 *
 */

int pvr2_send_sequence(struct pvr2_hdw *hdw, struct pvr2_cmd *cmd)
{
    int ret = 0, i;

    i = 0;
    while ((cmd[i].length != 0) && (ret == 0)){
        memcpy(hdw->cmd_buffer, cmd[i].cmd, cmd[i].length);
        ret |= pvr2_send_request(hdw, hdw->cmd_buffer, cmd[i].length,
                                 hdw->cmd_buffer, cmd[i].ret);
        i++;
    }
    return ret;
}

int pvr2_write_register(struct pvr2_hdw *hdw, u16 reg, u32 data)
{
    int ret;

    LOCK_TAKE(hdw->ctl_lock);

    hdw->cmd_buffer[0] = 0x04;  /* write register prefix */
    PVR2_DECOMPOSE_LE(hdw->cmd_buffer,1,data);
    hdw->cmd_buffer[5] = 0;
    hdw->cmd_buffer[6] = (reg >> 8) & 0xff;
    hdw->cmd_buffer[7] = reg & 0xff;


    ret = pvr2_send_request(hdw, hdw->cmd_buffer, 8, hdw->cmd_buffer, 0);

    LOCK_GIVE(hdw->ctl_lock);

    return ret;
}

int pvr2_read_register(struct pvr2_hdw *hdw, u16 reg, u32 *data)
{
    int ret = 0;

    LOCK_TAKE(hdw->ctl_lock);

    hdw->cmd_buffer[0] = 0x05;  /* read register prefix */
    hdw->cmd_buffer[1] = 0;
    hdw->cmd_buffer[2] = 0;
    hdw->cmd_buffer[3] = 0;
    hdw->cmd_buffer[4] = 0;
    hdw->cmd_buffer[5] = 0;
    hdw->cmd_buffer[6] = (reg >> 8) & 0xff;
    hdw->cmd_buffer[7] = reg & 0xff;

    ret |= pvr2_send_request(hdw, hdw->cmd_buffer, 8, hdw->cmd_buffer, 4);
    *data = PVR2_COMPOSE_LE(hdw->cmd_buffer,0);

    LOCK_GIVE(hdw->ctl_lock);

    return ret;
}

int pvr2_write_u16(struct pvr2_hdw *hdw, u16 data, int res)
{
    int ret;

    LOCK_TAKE(hdw->ctl_lock);

    hdw->cmd_buffer[0] = (data >> 8) & 0xff;
    hdw->cmd_buffer[1] = data & 0xff;

    ret = pvr2_send_request(hdw, hdw->cmd_buffer, 2, hdw->cmd_buffer, res);

    LOCK_GIVE(hdw->ctl_lock);

    return ret;
}

int pvr2_write_u8(struct pvr2_hdw *hdw, u8 data, int res)
{
    int ret;

    LOCK_TAKE(hdw->ctl_lock);

    hdw->cmd_buffer[0] = data;

    ret = pvr2_send_request(hdw, hdw->cmd_buffer, 1, hdw->cmd_buffer, res);

    LOCK_GIVE(hdw->ctl_lock);

    return ret;
}

/*
 * pvr_do_command()
 *
 * Executes a simple command.
 */

int pvr2_do_command(struct pvr2_hdw *hdw, struct pvr2_cmd *cmd)
{
    int ret;

    if (cmd == NULL)
        return -EIO;

    LOCK_TAKE(hdw->ctl_lock);

    /* data command (cmd->cmd) is (often) statically allocated. For usb
     *  transfer we need dynamic memory.
     */
    memcpy(hdw->cmd_buffer, cmd->cmd, cmd->length);

    ret = pvr2_send_request(hdw, hdw->cmd_buffer, cmd->length,
                            hdw->cmd_buffer, cmd->ret);

    LOCK_GIVE(hdw->ctl_lock);

    return ret;
}

void pvr2_hdw_render_useless_unlocked(struct pvr2_hdw *hdw)
{
    if (!hdw->flag_ok) return;
    pvr2_trace(PVR2_TRACE_INIT,"render_useless");
    hdw->flag_ok = 0;
    if (hdw->vid_stream) {
        pvr2_stream_setup(hdw->vid_stream,0,0);
    }
    hdw->flag_streaming_enabled = 0;
    hdw->subsys_enabled_mask = 0;
}


void pvr2_hdw_render_useless(struct pvr2_hdw *hdw)
{
    LOCK_TAKE(hdw->ctl_lock);
    pvr2_hdw_render_useless_unlocked(hdw);
    LOCK_GIVE(hdw->ctl_lock);
}


void pvr2_hdw_device_reset(struct pvr2_hdw *hdw)
{
    int ret;
    pvr2_trace(PVR2_TRACE_INIT,"Performing a device reset...");
    ret = usb_lock_device_for_reset(hdw->usb_dev,0);
    if (ret == 1) {
        ret = usb_reset_device(hdw->usb_dev);
        usb_unlock_device(hdw->usb_dev);
    } else {
        pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "Failed to lock USB device ret=%d",ret);
    }
}


void pvr2_hdw_cpureset_assert(struct pvr2_hdw *hdw,int val)
{
    char da[1];
    unsigned int pipe;
    int ret;

    if (!hdw->usb_dev) return;

    pvr2_trace(PVR2_TRACE_INIT,"cpureset_assert(%d)",val);

    da[0] = val ? 0x01 : 0x00;

    /* Write the CPUCS register on the 8051.  The lsb of the register is
       the reset bit; a 1 asserts reset while a 0 clears it. */
    pipe = usb_sndctrlpipe(hdw->usb_dev, 0);
    ret = usb_control_msg(hdw->usb_dev,pipe,0xa0,0x40,0xe600,0,da,1,HZ);
    if (ret < 0) {
        pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "cpureset_assert(%d) error=%d",val,ret);
        pvr2_hdw_render_useless(hdw);
    }
}


int pvr2_hdw_cmd_deep_reset(struct pvr2_hdw *hdw)
{
    int status;
    LOCK_TAKE(hdw->ctl_lock); do {
        pvr2_trace(PVR2_TRACE_INIT,"Requesting uproc hard reset");
        hdw->flag_ok = !0;
        hdw->cmd_buffer[0] = 0xdd;
        status = pvr2_send_request(hdw,hdw->cmd_buffer,1,0,0);
    } while (0); LOCK_GIVE(hdw->ctl_lock);
    return status;
}


int pvr2_hdw_cmd_soft_reset(struct pvr2_hdw *hdw)
{
    int status;
    LOCK_TAKE(hdw->ctl_lock); do {
        pvr2_trace(PVR2_TRACE_INIT,"Requesting uproc soft reset");
        hdw->cmd_buffer[0] = 0xde;
        status = pvr2_send_request(hdw,hdw->cmd_buffer,1,0,0);
    } while (0); LOCK_GIVE(hdw->ctl_lock);
    return status;
}


int pvr2_hdw_cmd_usbstream(struct pvr2_hdw *hdw,int runFl)
{
    int status;
    LOCK_TAKE(hdw->ctl_lock); do {
        hdw->cmd_buffer[0] = (runFl ? 0x36 : 0x37);
        status = pvr2_send_request(hdw,hdw->cmd_buffer,1,0,0);
    } while (0); LOCK_GIVE(hdw->ctl_lock);
    if (!status) {
        hdw->subsys_enabled_mask =
            ((hdw->subsys_enabled_mask & ~PVR2_SUBSYS_USBSTREAM_RUN) |
             (runFl ? PVR2_SUBSYS_USBSTREAM_RUN : 0));
    }
    return status;
}


void pvr2_hdw_get_debug_info(const struct pvr2_hdw *hdw,
                             struct pvr2_hdw_debug_info *ptr)
{
    ptr->big_lock_held = hdw->big_lock_held;
    ptr->ctl_lock_held = hdw->ctl_lock_held;
    ptr->flag_ok = hdw->flag_ok;
    ptr->flag_disconnected = hdw->flag_disconnected;
    ptr->flag_init_ok = hdw->flag_init_ok;
    ptr->flag_streaming_enabled = hdw->flag_streaming_enabled;
    ptr->subsys_flags = hdw->subsys_enabled_mask;
    ptr->cmd_debug_state = hdw->cmd_debug_state;
    ptr->cmd_code = hdw->cmd_debug_code;
    ptr->cmd_debug_write_len = hdw->cmd_debug_write_len;
    ptr->cmd_debug_read_len = hdw->cmd_debug_read_len;
    ptr->cmd_debug_timeout = hdw->ctl_timeout_flag;
    ptr->cmd_debug_write_pend = hdw->ctl_write_pend_flag;
    ptr->cmd_debug_read_pend = hdw->ctl_read_pend_flag;
    ptr->cmd_debug_rstatus = hdw->ctl_read_urb->status;
    ptr->cmd_debug_wstatus = hdw->ctl_read_urb->status;
}


#define PVR2_GPIO_IN 0x9008
#define PVR2_GPIO_OUT 0x900c
#define PVR2_GPIO_DIR 0x9020

int pvr2_hdw_gpio_get_dir(struct pvr2_hdw *hdw,u32 *dp)
{
    return pvr2_read_register(hdw,PVR2_GPIO_DIR,dp);
}


int pvr2_hdw_gpio_get_out(struct pvr2_hdw *hdw,u32 *dp)
{
    return pvr2_read_register(hdw,PVR2_GPIO_OUT,dp);
}


int pvr2_hdw_gpio_get_in(struct pvr2_hdw *hdw,u32 *dp)
{
    return pvr2_read_register(hdw,PVR2_GPIO_IN,dp);
}


int pvr2_hdw_gpio_chg_dir(struct pvr2_hdw *hdw,u32 msk,u32 val)
{
    u32 cval,nval;
    int ret;
    if (~msk) {
        ret = pvr2_read_register(hdw,PVR2_GPIO_DIR,&cval);
        if (ret) return ret;
        nval = (cval & ~msk) | (val & msk);
        pvr2_trace(PVR2_TRACE_GPIO,
                   "GPIO direction changing 0x%x:0x%x from 0x%x to 0x%x",
                   msk,val,cval,nval);
    } else {
        nval = val;
        pvr2_trace(PVR2_TRACE_GPIO,
                   "GPIO direction changing to 0x%x",nval);
    }
    return pvr2_write_register(hdw,PVR2_GPIO_DIR,nval);
}


int pvr2_hdw_gpio_chg_out(struct pvr2_hdw *hdw,u32 msk,u32 val)
{
    u32 cval,nval;
    int ret;
    if (~msk) {
        ret = pvr2_read_register(hdw,PVR2_GPIO_OUT,&cval);
        if (ret) return ret;
        nval = (cval & ~msk) | (val & msk);
        pvr2_trace(PVR2_TRACE_GPIO,
                   "GPIO output changing 0x%x:0x%x from 0x%x to 0x%x",
                   msk,val,cval,nval);
    } else {
        nval = val;
        pvr2_trace(PVR2_TRACE_GPIO,
                   "GPIO output changing to 0x%x",nval);
    }
    return pvr2_write_register(hdw,PVR2_GPIO_OUT,nval);
}



/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 75 ***
  *** c-basic-offset: 4 ***
  *** indent-tabs-mode: nil ***
  *** End: ***

  Similarly for vim to see:
  vim:expandtab:textwidth=75

*/
