/*
 *
 *  $Id: pvrusb2-iommap.c 496 2005-07-16 19:47:53Z isely $
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2-ioread.h"
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>

#define BUFFER_ORDER 2
#define BUFFER_SIZE PAGE_SIZE * (1 << BUFFER_ORDER)

struct pvr2_iommap {
    struct pvr2_stream *stream;
    char **buffer_storage;
    unsigned int buffer_count;
    struct semaphore sem;
};

struct pvr2_iommap *pvr2_iommap_create(struct pvr2_stream *sp,
                                       int buffer_count)
{
    unsigned int idx;
    struct pvr2_iommap *cp;
    struct pvr2_buffer *bp;
    cp = kmalloc(sizeof(*cp),GFP_KERNEL);
    memset(cp,0,sizeof(*cp));
    init_MUTEX(&cp->sem);
    cp->stream = sp;
    if (!buffer_count) buffer_count = 32;
    cp->buffer_storage = kmalloc(sizeof(char *) * buffer_count,GFP_KERNEL);
    if (!cp->buffer_storage) goto fail;
    memset(cp->buffer_storage,0,sizeof(char *) * buffer_count);
    cp->buffer_count = buffer_count;
    for (idx = 0; idx < cp->buffer_count; idx++) {
        cp->buffer_storage[idx] =
            (char *)__get_free_pages(GFP_KERNEL,BUFFER_ORDER);
        if (!cp->buffer_storage[idx]) goto fail;
        memset(cp->buffer_storage[idx],0,BUFFER_SIZE);
    }
    pvr2_stream_kill(cp->stream);
    if (pvr2_stream_set_buffer_count(sp,cp->buffer_count) < 0) goto fail;
    for (idx = 0; idx < buffer_count; idx++) {
        bp = pvr2_stream_get_buffer(sp,idx);
        pvr2_buffer_set_buffer(bp,cp->buffer_storage[idx],BUFFER_SIZE);
    }
    cp->stream = sp;
    return cp;

 fail:
    if (cp->buffer_storage) {
        for (idx = 0; idx < cp->buffer_count; idx++) {
            if (!(cp->buffer_storage[idx])) continue;
            free_pages((unsigned long)(cp->buffer_storage[idx]),BUFFER_ORDER);
            cp->buffer_storage[idx] = 0;
        }
        kfree(cp->buffer_storage);
    }
    kfree(cp);
    return 0;
}

void pvr2_iommap_destroy(struct pvr2_iommap *cp)
{
    unsigned int idx;
    if (cp->stream) {
        pvr2_stream_kill(cp->stream);
        pvr2_stream_set_buffer_count(cp->stream,0);
    }
    if (cp->buffer_storage) {
        for (idx = 0; idx < cp->buffer_count; idx++) {
            if (!(cp->buffer_storage[idx])) continue;
            free_pages((unsigned long)(cp->buffer_storage[idx]),BUFFER_ORDER);
            cp->buffer_storage[idx] = 0;
        }
        kfree(cp->buffer_storage);
    }
    cp->stream = 0;
    kfree(cp);
}

static void pvr2_iommap_vm_open(struct vm_area_struct *vma)
{
#ifdef notdef
    struct pvr2_iommap *cp = vma->vm_private_data;
    /* ????? */
#endif
}

static void pvr2_iommap_vm_close(struct vm_area_struct *vma)
{
#ifdef notdef
    struct pvr2_iommap *cp = vma->vm_private_data;
    /* ????? */
#endif
}

static struct page *pvr2_iommap_vm_nopage(struct vm_area_struct *vma,
                                          unsigned long address,int *type)
{
    struct pvr2_iommap *cp = vma->vm_private_data;
    struct page *page = NOPAGE_SIGBUS;
    unsigned long offset;
    unsigned long buffer_id;
    char *bp;
    down(&cp->sem);
    offset = (address - vma->vm_start) + (vma->vm_pgoff << PAGE_SHIFT);
    buffer_id = offset / BUFFER_SIZE;
    if (buffer_id < cp->buffer_count) {
        bp = cp->buffer_storage[buffer_id];
        bp += (offset % BUFFER_SIZE);
        page = virt_to_page(bp);
        if (page) {
            get_page(page);
            if (type) *type = VM_FAULT_MINOR;
        } else {
            page = NOPAGE_OOM;
        }
    }
    up(&cp->sem);
    return page;
}

static struct vm_operations_struct pvr2_iommap_vm_ops = {
    .open = pvr2_iommap_vm_open,
    .close = pvr2_iommap_vm_close,
    .nopage = pvr2_iommap_vm_nopage,
};

int pvr2_iommap_range(struct pvr2_iommap *cp,struct vm_area_struct *vma)
{
    down(&cp->sem);
    vma->vm_ops = &pvr2_iommap_vm_ops;
    vma->vm_flags |= VM_RESERVED | VM_IO;
    vma->vm_private_data = cp;
    up(&cp->sem);
    pvr2_iommap_vm_open(vma);
    return 0;
}


/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 75 ***
  *** c-basic-offset: 4 ***
  *** indent-tabs-mode: nil ***
  *** End: ***

  Similarly for vim to see:
  vim:expandtab:textwidth=75

*/
