#!/usr/bin/perl

# $Id: fwextract.pl 525 2005-07-23 02:24:50Z isely $

#
#  Copyright (C) 2005 Mike Isely <isely@pobox.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

# This script automates the extraction of firmware from the Hauppauge
# Windows driver files.  This script can also be used to assist in
# locating newer firmware in newer driver files.

# To extract firmware, just run the script, ensuring that the Windows
# driver files have been expanded into a "win_driver" subdirectory.
# (Alternatively, use the --driverDir option to specify where the
# driver files are.)  The script will locate the firmware and extract
# it to two files, pvrusb2.f1 and pvrusb2.f2 which will be ready to
# use by the pvrusb2 driver.  This script uses a table to recognize
# firmware files; it is able to seek and recognize multiple driver
# versions (depending on what's in the table).

# To teach the script about newer driver versions, you must first find
# an alternate way to manually extract the firmware files.  This can
# be accomplished either by manual hacking (as others have been doing
# already), or you can use additional tools in the pvrusb2 driver
# package to suck the firmware back out of the hardware.  In either
# case, once you have the two firmware files manually extracted (and
# named pvrusb2.f1 and pvrusb2.f2 in the current directory), just run
# this script with the --search option.  This will cause it to use
# those two firmware files as search keys into the Windows drivers.
# It will search through the driver files and find correct offsets and
# sizes for where the firmware has been "hidden".  It will then print
# out a Perl snippet that you can append to the $contentInfo structure
# further below.  Send that snippet back to me (Mike Isely
# <isely@pobox.com>) and I'll put it back into the master copy of this
# script.  Then others will be able to easily extract the newer
# firmware without having to go through the manual pain you will have
# just completed :-)

use strict;
use Getopt::Long;
use Digest::MD5 qw(md5 md5_hex md5_base64);
use IO::File;
use IO::Dir;


# -------------------------------------------------------------------------

# Default values for the Windows driver file directory and the names
# for the two firmware files to be extracted / searched.  Command line
# options may be used to change these (look at the end of this file -
# we're just using Getopt to do the dirty work for that).

my $driverDir = "win_driver";
my $f1File = "pvrusb2.f1";
my $f2File = "pvrusb2.f2";


# -------------------------------------------------------------------------

# The table below tells the extraction process where to find the
# firmware data.  What the extraction process does is blindly MD5 sum
# each windows driver file and try to look that key up here.  (This is
# an extremely fast operation).  If the lookup succeeds, then we know
# we've located a file with something interesting in it.  What's
# "interesting" is described by the table entry, keyed by "fw1" and
# "fw2".  For each key, the nested data describes the offset and size
# of the data to extract and also includes another MD5 sum to ensure
# that the extraction was truly successful.

# Note that using MD5 sums effectively as the search key, we are
# completely independent of the driver file name and driver package
# version in which it was found.  To make the extractor work with
# additional driver versions, one need only add entries to this table
# - the extractor will do the rest.

# To add entries to this table you'll still have to hand edit this
# script, however you don't have to guess at the data.  Run this
# script with the --search option and manually-extracted versions of
# the firmware files nearby and the script will execute an exhaustive
# / fuzzy search to locate the embedded firmware.  Then it will print
# the resulting table entry that you must add below.

my $contentInfo = {
  bf1cc4da051b1db04091562c86c691ff => {
    comment => "pvrusb2_23_31351",
    fw1 => {
      md5_hex => "3b7288416144467369ed2b5b10cf931d",
      offset => "987584",
      size => "8192"
    },
    fw2 => {
      md5_hex => "ab75947ef1b086e26f9b08e628baa02e",
      offset => "201152",
      size => "262144"
    }
  },
  "02e4934b8525db1fee074f57a0c27304" => {
    comment => 'CDROM 2.5.22329 (credit: Eliette Mathey <elmat52@yahoo.fr>)',
    fw1 => {
      md5_hex => "3b7288416144467369ed2b5b10cf931d",
      offset => "1065472",
      size => "8192"
    },
    fw2 => {
      md5_hex => "79c5daf4cde87036c834a314b4929fb1",
      offset => "279040",
      size => "262144"
    }
  },
  b100615d9497d205b065a985bc9d73c3 => {
    comment => "pvrusb2_inf_26c (related to pvrusb2_26_23097)",
    fw1 => {
      md5_hex => "3b7288416144467369ed2b5b10cf931d",
      offset => "1072928",
      size => "8192"
    },
    fw2 => {
      md5_hex => "79c5daf4cde87036c834a314b4929fb1",
      offset => "286496",
      size => "262144"
    }
  },
  "6576e6cc206a4b5469345e7ebe578abf" => {
    comment => "pvrusb2_inf (related to pvrusb2_26_23055)",
    fw1 => {
      md5_hex => "3b7288416144467369ed2b5b10cf931d",
      offset => "1072992",
      size => "8192"
    },
    fw2 => {
      md5_hex => "79c5daf4cde87036c834a314b4929fb1",
      offset => "286560",
      size => "262144"
    }
  },
  "9c5fc27cd3a5c2307c911cad92c95689" => {
    comment => "pvrusb2_27_23178",
    fw1 => {
      md5_hex => "3b7288416144467369ed2b5b10cf931d",
      offset => "1073216",
      size => "8192"
    },
    fw2 => {
      md5_hex => "79c5daf4cde87036c834a314b4929fb1",
      offset => "286784",
      size => "262144"
    }
  },
};


# -------------------------------------------------------------------------

# This is all utility code that I've ripped from other code that I've
# written...

# This transforms the string such that it can be safely printed
# between double-quotes and then eval'ed back in at a later time to
# produce the original string value:
#  - Escape double quotes
#  - Escape dollar sign
#  - Escape '@' sign
#  - Escape newline
#  - Escape backslash
#  - Hexify any non-printing character
sub perlQuoteStr {
    my $str = shift;

    $str =~ s/([\$\@"\\])/\\$1/g;
    $str =~ s/([\n])/\\n/g;
    $str =~ s{([\x00-\x1f\x80-\xff])}{sprintf('\\x%02x',ord($1))}ge;

    return $str;
}

# Return string quoted or bare if it's safe to represent the string as
# a bareword.
sub perlSafeEmitBareStr {
    my $str = shift;
    # Two tests here: First only strings composed entirely of letters,
    # digits and the underscore - with the first character being a
    # letter - passes.  Then fail any string which might be mistaken
    # for a "v-string" (yuck).  If we fail, we quote the whole thing.
    # Otherwise we'll just return the bareword.
    if (($str =~ /^[a-zA-Z]\w+$/) && !($str =~ /^v\d+$/)) {
	return $str;
    }
    return '"' . perlQuoteStr($str) . '"';
}

# Return string quoted or bare if it is just a single digit.
sub perlSafeEmitStr {
    my $str = shift;
    if ($str =~ /^\d$/) {
	# Single-digit case - easily handled
	return $str;
    }
    return '"' . perlQuoteStr($str) . '"';
}

sub structureWriteToHandle2 {
    my $fh = shift;
    my $var = shift;
    my $prefix = shift;
    my ($v1,$v2);

    if (! defined $var) {
	print $fh 'undef';
	return;
    }
    $v2 = ref $var;
    if ($v2 eq "") {
	print $fh perlSafeEmitStr($var);
	return;
    }
    if ($v2 eq 'ARRAY') {
	my $oldPrefix = $prefix;
	$prefix = $prefix . "  ";
	print $fh "[\n";
	for ($v1 = 0; $v1 < scalar(@$var); $v1++) {
	    if ($v1) {
		print $fh ",\n";
	    }
	    print $fh $prefix;
	    structureWriteToHandle2($fh,$var->[$v1],$prefix);
	}
	print $fh "\n";
	print $fh $oldPrefix;
	print $fh "]";
	return;
    }
    if ($v2 eq 'HASH') {
	my $key;
	my $oldPrefix = $prefix;
	$prefix = $prefix . "  ";
	print $fh "{\n";
	$v1 = 0;
	foreach $key (sort keys %$var) {
	    if ($v1) {
		print $fh ",\n";
	    }
	    print $fh $prefix;
	    print $fh perlSafeEmitBareStr($key) . ' => ';
	    structureWriteToHandle2($fh,$var->{$key},$prefix);
	    $v1++;
	}
	print $fh "\n";
	print $fh $oldPrefix;
	print $fh "}";
	return;
    }

    return;
}

sub structureWriteToHandle {
    my $fh = shift;
    my $var = shift;
    my $prefix = shift;

    structureWriteToHandle2($fh,$var,$prefix);
    print $fh "\n";
}


# -------------------------------------------------------------------------

# Suck a file into a data structure so that we can manipulate it.
sub fetchFileData {
    my $fname = shift;
    my $fh = new IO::File;
    my $cnt;

    my $info = {
	name => $fname,
	data => '',
    };

    die "Failed to open $fname" unless ($fh->open("<$fname"));

    while ($fh->read($info->{data},16384,length $info->{data})) { }

    $info->{size} = length $info->{data};
    $info->{md5_hex} = md5_hex($info->{data});
    return $info;
};

# Print a <value> / <value> string to the terminal.  We do this so
# that the user doesn't thing that the script is hung.   This is a
# helper for searchForFuzzySequence.
sub printProgress {
    my $ctxt = shift;
    my $idx = shift;
    my $str;

    print "\r" if (defined $ctxt->{dirty});
    $ctxt->{dirty} = !0;
    $str = sprintf "%7d / %7d",$idx,$ctxt->{cnt};
    print "$str";
}

# Clean up after printing a <value> / <value> string to the terminal.
# We do this so that the user doesn't thing that the script is hung.
# This is a helper for searchForFuzzySequence.
sub clearProgress {
    my $ctxt = shift;
    my $idx = shift;
    return if (! defined $ctxt->{dirty});
    printProgress($ctxt,$idx);
    print "\n";
}

# Search for a needle in a haystack.  Pass in the haystack and the
# needle and try to find where in the haystack it is a residing.
# Return value is the offset within the haystack where it's been found
# or undef.  This implementation is a slow "fuzzy" search.  We
# tolerate up to 1% of the data bytes being incorrect, which allows us
# to use slightly changed firmware data sucked from the hardware as a
# search key (the 8051 firmware in fact seems to always come back with
# a single modified byte).
sub searchForFuzzySequence {
    my $needleInfo = shift;
    my $haystackInfo = shift;
    my ($idx1,$idx2);
    my $errCnt;
    my $errMaxCnt;
    my $needleCheckSize;
    my $needleRef = \$needleInfo->{data};
    my $haystackRef = \$haystackInfo->{data};
    my $needleSize = length $$needleRef;
    my $haystackSize = length $$haystackRef;
    my $pctxt = { };

    return undef if ($haystackSize < $needleSize);

    my $range = $haystackSize - $needleSize;

    $errMaxCnt = int($needleSize * 0.002);
    $needleCheckSize = int($needleSize * 0.75);

#    print "Fuzzy check info:\n";
#    print "Checking $needleCheckSize out of $needleSize bytes\n";
#    print "Will tolerate up to $errMaxCnt errors\n";

    $pctxt->{cnt} = $range;
    for ($idx1 = 0; $idx1 < $range; $idx1++) {
	if ($idx1 && !($idx1 % 10000)) {
	    printProgress($pctxt,$idx1);
	}
	$errCnt = 0;
	for ($idx2 = 0; $idx2 < $needleCheckSize; $idx2++) {
	    next if (substr($$haystackRef,$idx1+$idx2,1) eq
		     substr($$needleRef,$idx2,1));
	    $errCnt++;
	    last if ($errCnt > $errMaxCnt);
	}
	next if ($idx2 < $needleCheckSize);
	clearProgress($pctxt,$idx1);
	return $idx1;
    }
    clearProgress($pctxt,$idx1);
    return undef;
};

# Search for a needle in a haystack.  Pass in the haystack and the
# needle and try to find where in the haystack it is a residing.
# Return value is the offset within the haystack where it's been found
# or undef.  This is a fast, exact search.
sub searchForSequence {
    my $needleInfo = shift;
    my $haystackInfo = shift;
    my $idx;
    my $needleSize = length $needleInfo->{data};
    my $haystackSize = length $haystackInfo->{data};

    return undef if ($haystackSize < $needleSize);

    my $range = $haystackSize - $needleSize;
    my $ssize = 4;

    # Pure paranoia
    $ssize = $needleSize if ($ssize > $needleSize);

    for ($idx = 0; $idx < $range; $idx++) {
	# Just check the first 4 bytes initially since this will be fast
	next if (substr($haystackInfo->{data},$idx,$ssize) ne
		 substr($needleInfo->{data},0,$ssize));
	# If we get here, then it's probably a match, but let's check
	# the entire data block now.
	next if (substr($haystackInfo->{data},$idx,$needleSize) ne
		 $needleInfo->{data});
	return $idx;
    }
    return undef;
};

# Helper function for recording info in a data structure about
# firmware we've found.
sub noteResult {
    my $result = shift;
    my $fileInfo = shift;
    my $fwInfo = shift;

    $result->{$fileInfo->{md5_hex}}->{$fwInfo->{type}} = {
	offset => $fwInfo->{offset},
	size => $fwInfo->{size},
	md5_hex => md5_hex(substr($fileInfo->{data},
				  $fwInfo->{offset},
				  $fwInfo->{size})),
    };
}

# Helper function to search a single file for firmware.  This only
# does an exact search.
sub searchASingleFile {
    my $fileData = shift;
    my $fw1Info = shift;
    my $fw2Info = shift;
    my $offset;

    if ($fw1Info) {
	$offset = searchForSequence($fw1Info,$fileData);
	if (defined $offset) {
	    $fw1Info->{offset} = $offset;
	}
    }

    if ($fw2Info) {
	$offset = searchForSequence($fw2Info,$fileData);
	if (defined $offset) {
	    $fw2Info->{offset} = $offset;
	}
    }
}

# Quick and dirty function to grab the contents of a directory.
sub getDirEntries {
    my $dirName = shift;
    my $dirh;
    my @entries;

    $dirh = new IO::Dir $dirName;

    return undef if (! defined $dirh);

    @entries = sort $dirh->read();

    return \@entries;
}

# This logic assumes that the two firmware files already exist and
# attempts to locate where the data has been embedded within the windows
# driver files.  All files are searched, and if we find the embedded
# data, we'll print out a Perl data structure that can be inserted
# into this program's $contentInfo structure which will tell this
# program how to later extract the firmware files from those same places.
sub searchDriver {
    my $comment = shift;
    my $fw1Info;
    my $fw2Info;
    my $fileInfo;
    my ($v1,$v2,$v3);
    my $result = {
    };

    # First read in the already expected two firmware image files.
    if (defined $f1File) {
	$fw1Info = fetchFileData($f1File);
	$fw1Info->{type} = 'fw1';
    }
    if (defined $f2File) {
	$fw2Info = fetchFileData($f2File);
	$fw2Info->{type} = 'fw2';
    }

    # Grab set of files we're going to scan
    my $files = getDirEntries($driverDir);

    die "Failed to read windows driver directory $driverDir"
	if not defined($files);

    # Look in each file for our expected firmware, using an exact
    # search.  This is fastest.
    print "Performing exact search in $driverDir\n";
    foreach $v1 (@$files) {
	next if ($v1 eq '.');
	next if ($v1 eq '..');
	last if (!((defined $fw1Info) || (defined $fw2Info)));
	$v2 = $driverDir . '/' . $v1;
	$fileInfo = fetchFileData($v2);
	print "Searching $fileInfo->{name} (size=$fileInfo->{size})\n";
	searchASingleFile($fileInfo,$fw1Info,$fw2Info);
	if (($fw1Info && exists $fw1Info->{offset}) ||
	    ($fw2Info && exists $fw2Info->{offset})) {
	    # Found something
	    $result->{$fileInfo->{md5_hex}}->{comment} = $comment;
	    if ($fw1Info && exists $fw1Info->{offset}) {
		# Found fw1 data
		noteResult($result,$fileInfo,$fw1Info);
		$fw1Info = undef;
	    }
	    if ($fw2Info && exists $fw2Info->{offset}) {
		# Found fw2 data
		noteResult($result,$fileInfo,$fw2Info);
		$fw2Info = undef;
	    }
	}
    }

    if (defined $fw1Info) {
	print "Performing fuzzy search in $driverDir for $fw1Info->{type}\n";
	# First look in files where we had previously found something
	foreach $v1 (@$files) {
	    next if ($v1 eq '.');
	    next if ($v1 eq '..');
	    last if (!(defined $fw1Info));
	    $v2 = $driverDir . '/' . $v1;
	    $fileInfo = fetchFileData($v2);
	    next if (!exists $result->{$fileInfo->{md5_hex}});
	    print "Fuzzy searching $fileInfo->{name}" .
		" (size=$fileInfo->{size})\n";
	    $v3 = searchForFuzzySequence($fw1Info,$fileInfo);
	    next if (! defined $v3);
	    $fw1Info->{offset} = $v3;
	    noteResult($result,$fileInfo,$fw1Info);
	    $fw1Info = undef;
	    last;
	}
	if (defined $fw1Info) {
	    # Now look in files where we had not previously found something
	    foreach $v1 (@$files) {
		next if ($v1 eq '.');
		next if ($v1 eq '..');
		last if (!(defined $fw1Info));
		$v2 = $driverDir . '/' . $v1;
		$fileInfo = fetchFileData($v2);
		next if (exists $result->{$fileInfo->{md5_hex}});
		print "Fuzzy searching $fileInfo->{name}" .
		    " (size=$fileInfo->{size})\n";
		$v3 = searchForFuzzySequence($fw1Info,$fileInfo);
		next if (! defined $v3);
		$fw1Info->{offset} = $v3;
		noteResult($result,$fileInfo,$fw1Info);
		$fw1Info = undef;
		last;
	    }
	}
    }

    if (defined $fw2Info) {
	print "Performing fuzzy search in $driverDir for $fw2Info->{type}\n";
	# First look in files where we had previously found something
	foreach $v1 (@$files) {
	    next if ($v1 eq '.');
	    next if ($v1 eq '..');
	    last if (!(defined $fw2Info));
	    $v2 = $driverDir . '/' . $v1;
	    $fileInfo = fetchFileData($v2);
	    print "Fuzzy searching $fileInfo->{name}" .
		" (size=$fileInfo->{size})\n";
	    next if (!exists $result->{$fileInfo->{md5_hex}});
	    $v3 = searchForFuzzySequence($fw2Info,$fileInfo);
	    next if (! defined $v3);
	    $fw2Info->{offset} = $v3;
	    noteResult($result,$fileInfo,$fw2Info);
	    $fw2Info = undef;
	    last;
	}
	if (defined $fw2Info) {
	    # Now look in files where we had not previously found something
	    foreach $v1 (@$files) {
		next if ($v1 eq '.');
		next if ($v1 eq '..');
		last if (!(defined $fw2Info));
		$v2 = $driverDir . '/' . $v1;
		$fileInfo = fetchFileData($v2);
		print "Fuzzy searching $fileInfo->{name}" .
		    " (size=$fileInfo->{size})\n";
		next if (exists $result->{$fileInfo->{md5_hex}});
		$v3 = searchForFuzzySequence($fw2Info,$fileInfo);
		next if (! defined $v3);
		$fw2Info->{offset} = $v3;
		noteResult($result,$fileInfo,$fw2Info);
		$fw2Info = undef;
		last;
	    }
	}
    }

    # Print out results...
    if (scalar(keys %$result)) {
	print 'Please add this fragment info the $contentInfo' .
	    ' structure of this program:';
	print "\n";
	structureWriteToHandle(\*STDOUT,$result);
	print <<EOF;
I also strongly encourage you to send me (Mike Isely <isely\@pobox.com>) the
above snippet so that I can update the master copy of this script and
others can benefit from your effort.
EOF
	return 0;
    } else {
	print "Failed to find firmware data within Windows driver files\n";
	return 1;
    }
}

# Extract firmware from parent file
sub doSubExtract {
    my $fileInfo = shift;
    my $fwContentInfo = shift;
    my $fwKey = shift;
    my $fwName = shift;
    my $fwInfo;
    my $fc;

    return undef if (! defined $fwContentInfo->{$fwKey});

    $fc = $fwContentInfo->{$fwKey};

    $fwInfo = {
	name => $fwName,
	parent => $fileInfo->{name},
	comment => $fwContentInfo->{comment},
	md5_hex => $fc->{md5_hex},
	size => $fc->{size},
	offset => $fc->{offset},
	data => substr($fileInfo->{data},
		       $fc->{offset},
		       $fc->{size}),
    };

    return $fwInfo;
}

# Write firmware to a file and report about what we've written.
sub doSubWriteback {
    my $fwInfo = shift;
    my $fname = shift;
    my $fh = new IO::File;
    my $md5_info;
    my $retCode = undef;

    if (!defined $fwInfo) {
	print STDERR "Error: Failed to locate data for writing $fname\n";
	return $retCode;
    }

    die "Failed to open $fname for writing" unless $fh->open(">$fname");

    print $fh $fwInfo->{data};

    $fh->close();

    print "Wrote $fname:\n";
    print "  Size $fwInfo->{size} bytes\n";
    print "  Driver package $fwInfo->{comment}\n";
    print "  Source $fwInfo->{parent} offset $fwInfo->{offset}\n";
    $md5_info = md5_hex($fwInfo->{data});
    if ($md5_info ne $fwInfo->{md5_hex}) {
	print "  Warning: MD5 mismatch on extracted content for $fname\n";
	print "  MD5 Expected: $fwInfo->{md5_hex}\n";
	print "  MD5      Got: $md5_info\n";
    } else {
	print "  MD5 verified ($md5_info)\n";
	$retCode = 1;
    }

    return $retCode;
}

# This scans a directory of Windows driver files and attempts to
# recognize files which contain firmware.  For those it finds, it
# executes the extraction process.
sub doExtract {
    my $fileInfo;
    my $fw1Info;
    my $fw2Info;
    my ($v1,$v2,$v3);
    my ($rc1,$rc2);

    # Grab set of files we're going to scan
    my $files = getDirEntries($driverDir);

    die "Failed to read windows driver directory $driverDir"
	if not defined($files);

    # Look in each file for our expected firmware
    foreach $v1 (@$files) {
	next if ($v1 eq '.');
	next if ($v1 eq '..');
	last if ((defined $fw1Info) && (defined $fw2Info));

	$v2 = $driverDir . '/' . $v1;
	$fileInfo = fetchFileData($v2);

	$fileInfo->{md5_hex} = md5_hex($fileInfo->{data});
	next if (! exists $contentInfo->{$fileInfo->{md5_hex}});

	print "Found firmware data in $v2\n";

	$v3 = $contentInfo->{$fileInfo->{md5_hex}};

	if (! defined $fw1Info) {
	    $fw1Info = doSubExtract($fileInfo,$v3,'fw1',$f1File);
	}
	if (! defined $fw2Info) {
	    $fw2Info = doSubExtract($fileInfo,$v3,'fw2',$f2File);
	}
    }

    $rc1 = doSubWriteback($fw1Info,$f1File);
    $rc2 = doSubWriteback($fw2Info,$f2File);

    return 0 if ((defined $rc1) && (defined $rc2));
    return 1;
}

# Our "main"
sub mainFunc {
    my $searchMode = 0;
    my $retCode;
    my $comment;

    exit 1 if (!GetOptions("fw1File=s" => \$f1File,
			   "fw2File=s" => \$f2File,
			   "comment=s" => \$comment,
			   "driverDir=s" => \$driverDir,
			   "search!" => \$searchMode));
    print "driverDir is $driverDir\n";
    print "fw1File is $f1File\n";
    print "fw2File is $f2File\n";

    if ($searchMode) {
	print "Searching for firmware images in driver directory\n";
	$retCode = searchDriver($comment);
    } else {
	print "Extracting firmware files from driver directory\n";
	$retCode = doExtract();
    }

    exit $retCode;
}

# Don't autoflush stdout so that progress printing works OK.
autoflush STDOUT !0;

mainFunc();
